/* 
 *	HT Editor
 *	textedit.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __TEXTEDIT_H__
#define __TEXTEDIT_H__

#include "htobj.h"
#include "textfile.h"

#define TEXTEDITOPT_NULL			0
#define TEXTEDITOPT_READONLY		1
#define TEXTEDITOPT_INPUTTABS		2

/*
 *	CLASS ht_rowcol_indicator
 */

struct ht_rowcol_indicator_data {
	dword row;
	dword col;
};
 
class ht_rowcol_indicator: public ht_view {
protected:
	UINT row;
	UINT col;
public:
			void init(bounds *b);
/* overwritten */
	virtual	void draw();
	virtual	void getdata(ht_object_stream *s);
	virtual	void setdata(ht_object_stream *s);
};

/*
 *	CLASS ht_text_editor
 */

struct text_editor_pos {
	UINT line;
	UINT pofs;
};

class ht_text_editor: public ht_view {
protected:
	ht_textfile *textfile;
	bool own_textfile;
	
	ht_syntax_lexer *lexer;
	bool own_lexer;

	UINT edit_options;

	UINT cursorx, cursory;
	text_editor_pos sel_start, sel_end;

	UINT top_line;
	UINT xofs;

	bool selectmode;
	text_editor_pos selectstart;

	ht_rowcol_indicator *rowcol_indicator;
/* new */
			UINT char_vsize(char c, UINT x);
			void clipboard_copy_cmd();
			void clipboard_paste_cmd();
			void clipboard_cut_cmd();
			void concat_lines(UINT a);
			UINT cursor_up(UINT n);
			UINT cursor_down(UINT n);
			UINT cursor_left(UINT n);
			UINT cursor_right(UINT n);
			void cursor_home();
			void cursor_end();
			void cursor_vput(UINT vx);
			void cursor_pput(UINT px);
			void delete_chars(UINT line, UINT ofs, UINT count);
			void delete_lines(UINT line, UINT count);
			UINT getline_vlength(UINT line);
			void insert_chars(UINT line, UINT ofs, void *chars, UINT len);
			void insert_lines(UINT line, UINT count);
			void make_pos_physical(text_editor_pos *p);
			void notify_cursor_changed();
			void normalize_selection();
			UINT physical_cursorx();
			void render_str(int x, int y, vcp color, text_editor_pos *pos, UINT len, char *str, bool multi);
			void render_str_color(vcp *color, text_editor_pos *pos);
			UINT scroll_up(UINT n);
			UINT scroll_down(UINT n);
			UINT scroll_left(UINT n);
			UINT scroll_right(UINT n);
			void select_add(text_editor_pos *from, text_editor_pos *to);
			void select_end();
			void select_set(text_editor_pos *from, text_editor_pos *to);
			void select_start();
			void split_line(UINT a, UINT pos);
public:
			void	init(bounds *b, bool own_textfile, ht_textfile *textfile, bool own_lexer, ht_syntax_lexer *lexer, UINT edit_options);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual 	void handlemsg(htmsg *msg);
	virtual	void resize(int rw, int rh);
/* new */
			void set_rowcol_indicator(ht_rowcol_indicator *r);
};

#endif /* __TEXTEDIT_H__ */

